@@ -9,8 +9,11 @@
 #include <config.h>
 #endif
 
+#include <gio/gio.h>
+
 #include "soup-client-input-stream.h"
-#include "soup.h"
+#include "soup-marshal.h"
+#include "soup-message.h"
 #include "soup-message-private.h"
 
 struct _SoupClientInputStreamPrivate {
@@ -18,7 +21,7 @@ struct _SoupClientInputStreamPrivate {
 };
 
 enum {
-	SIGNAL_EOF,
+	EOF,
 	LAST_SIGNAL
 };
 
@@ -34,29 +37,20 @@ static GPollableInputStreamInterface *soup_client_input_stream_parent_pollable_i
 static void soup_client_input_stream_pollable_init (GPollableInputStreamInterface *pollable_interface, gpointer interface_data);
 
 G_DEFINE_TYPE_WITH_CODE (SoupClientInputStream, soup_client_input_stream, SOUP_TYPE_FILTER_INPUT_STREAM,
-                         G_ADD_PRIVATE (SoupClientInputStream)
 			 G_IMPLEMENT_INTERFACE (G_TYPE_POLLABLE_INPUT_STREAM,
 						soup_client_input_stream_pollable_init))
 
 static void
 soup_client_input_stream_init (SoupClientInputStream *stream)
 {
-	stream->priv = soup_client_input_stream_get_instance_private (stream);
+	stream->priv = G_TYPE_INSTANCE_GET_PRIVATE (stream,
+						    SOUP_TYPE_CLIENT_INPUT_STREAM,
+						    SoupClientInputStreamPrivate);
 }
 
 static void
-soup_client_input_stream_finalize (GObject *object)
-{
-	SoupClientInputStream *cistream = SOUP_CLIENT_INPUT_STREAM (object);
-
-	g_clear_object (&cistream->priv->msg);
-
-	G_OBJECT_CLASS (soup_client_input_stream_parent_class)->finalize (object);
-}
-
-static void
-soup_client_input_stream_set_property (GObject *object, guint prop_id,
-				       const GValue *value, GParamSpec *pspec)
+set_property (GObject *object, guint prop_id,
+	      const GValue *value, GParamSpec *pspec)
 {
 	SoupClientInputStream *cistream = SOUP_CLIENT_INPUT_STREAM (object);
 
@@ -71,8 +65,8 @@ soup_client_input_stream_set_property (GObject *object, guint prop_id,
 }
 
 static void
-soup_client_input_stream_get_property (GObject *object, guint prop_id,
-				       GValue *value, GParamSpec *pspec)
+get_property (GObject *object, guint prop_id,
+	      GValue *value, GParamSpec *pspec)
 {
 	SoupClientInputStream *cistream = SOUP_CLIENT_INPUT_STREAM (object);
 
@@ -86,6 +80,17 @@ soup_client_input_stream_get_property (GObject *object, guint prop_id,
 	}
 }
 
+/* Temporary HACK to keep SoupCache working. See soup_client_input_stream_read_fn()
+ * and soup_client_input_stream_read_nonblocking().
+ */
+static void
+soup_client_input_stream_emit_got_chunk (SoupClientInputStream *stream, void *data, gssize nread)
+{
+	SoupBuffer *buffer = soup_buffer_new (SOUP_MEMORY_TEMPORARY, data, nread);
+	soup_message_got_chunk (stream->priv->msg, buffer);
+	soup_buffer_free (buffer);
+}
+
 static gssize
 soup_client_input_stream_read_fn (GInputStream  *stream,
 				  void          *buffer,
@@ -99,7 +104,13 @@ soup_client_input_stream_read_fn (GInputStream  *stream,
 		read_fn (stream, buffer, count, cancellable, error);
 
 	if (nread == 0)
-		g_signal_emit (stream, signals[SIGNAL_EOF], 0);
+		g_signal_emit (stream, signals[EOF], 0);
+
+	/* Temporary HACK to keep SoupCache working */
+	if (nread > 0) {
+		soup_client_input_stream_emit_got_chunk (SOUP_CLIENT_INPUT_STREAM (stream),
+							 buffer, nread);
+	}
 
 	return nread;
 }
@@ -116,7 +127,13 @@ soup_client_input_stream_read_nonblocking (GPollableInputStream  *stream,
 		read_nonblocking (stream, buffer, count, error);
 
 	if (nread == 0)
-		g_signal_emit (stream, signals[SIGNAL_EOF], 0);
+		g_signal_emit (stream, signals[EOF], 0);
+
+	/* Temporary HACK to keep SoupCache working */
+	if (nread > 0) {
+		soup_client_input_stream_emit_got_chunk (SOUP_CLIENT_INPUT_STREAM (stream),
+							 buffer, nread);
+	}
 
 	return nread;
 }
@@ -127,54 +144,69 @@ soup_client_input_stream_close_fn (GInputStream  *stream,
 				   GError       **error)
 {
 	SoupClientInputStream *cistream = SOUP_CLIENT_INPUT_STREAM (stream);
-	gboolean success;
 
-	success = soup_message_io_run_until_finish (cistream->priv->msg, TRUE,
-						    NULL, error);
-	soup_message_io_finished (cistream->priv->msg);
-	return success;
+	if (!soup_message_io_run_until_finish (cistream->priv->msg,
+					       cancellable, error))
+		return FALSE;
+
+	return G_INPUT_STREAM_CLASS (soup_client_input_stream_parent_class)->close_fn (stream, cancellable, error);
 }
 
-static gboolean
-idle_finish_close (gpointer user_data)
-{
-	GTask *task = user_data;
+typedef struct {
+	SoupClientInputStream *cistream;
+	gint priority;
+	GCancellable *cancellable;
+	GSimpleAsyncResult *result;
+} CloseAsyncData;
 
-	g_task_return_boolean (task, TRUE);
-	g_object_unref (task);
-	return FALSE;
+static void
+close_async_data_free (CloseAsyncData *cad)
+{
+	if (cad->cancellable)
+		g_object_unref (cad->cancellable);
+	g_object_unref (cad->result);
+	g_slice_free (CloseAsyncData, cad);
+}
+
+static void
+base_stream_closed (GObject *source, GAsyncResult *result, gpointer user_data)
+{
+	CloseAsyncData *cad = user_data;
+	GError *error = NULL;
+
+	if (G_INPUT_STREAM_CLASS (soup_client_input_stream_parent_class)->
+	    close_finish (G_INPUT_STREAM (cad->cistream), result, &error))
+		g_simple_async_result_set_op_res_gboolean (cad->result, TRUE);
+	else
+		g_simple_async_result_take_error (cad->result, error);
+
+	g_simple_async_result_complete_in_idle (cad->result);
+	close_async_data_free (cad);
 }
 
 static gboolean
 close_async_ready (SoupMessage *msg, gpointer user_data)
 {
-	GTask *task = user_data;
-	SoupClientInputStream *cistream = g_task_get_source_object (task);
+	CloseAsyncData *cad = user_data;
 	GError *error = NULL;
 
-	if (!soup_message_io_run_until_finish (cistream->priv->msg, FALSE,
-					       g_task_get_cancellable (task),
-					       &error) &&
-	    g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
-		g_error_free (error);
-		return TRUE;
-	}
-
-	soup_message_io_finished (cistream->priv->msg);
-
-	if (error) {
-		g_task_return_error (task, error);
-		g_object_unref (task);
+	if (soup_message_io_run_until_finish (cad->cistream->priv->msg,
+					      cad->cancellable, &error)) {
+		G_INPUT_STREAM_CLASS (soup_client_input_stream_parent_class)->
+			close_async (G_INPUT_STREAM (cad->cistream),
+				     cad->priority,
+				     cad->cancellable,
+				     base_stream_closed,
+				     cad);
+		return FALSE;
+	} else if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
+		g_simple_async_result_take_error (cad->result, error);
+		g_simple_async_result_complete_in_idle (cad->result);
+		close_async_data_free (cad);
 		return FALSE;
 	}
 
-	/* Due to a historical accident, SoupSessionAsync relies on us
-	 * waiting one extra cycle after run_until_finish() returns.
-	 * Ugh. FIXME later when it's easier to do.
-	 */
-	soup_add_idle (g_main_context_get_thread_default (),
-		       idle_finish_close, task);
-	return FALSE;
+	return TRUE;
 }
 
 static void
@@ -184,20 +216,23 @@ soup_client_input_stream_close_async (GInputStream        *stream,
 				      GAsyncReadyCallback  callback,
 				      gpointer             user_data)
 {
-	SoupClientInputStream *cistream = SOUP_CLIENT_INPUT_STREAM (stream);
-	GTask *task;
+	CloseAsyncData *cad;
 	GSource *source;
 
-	task = g_task_new (stream, cancellable, callback, user_data);
-	g_task_set_priority (task, priority);
+	cad = g_slice_new (CloseAsyncData);
+	cad->cistream = SOUP_CLIENT_INPUT_STREAM (stream);
+	cad->result = g_simple_async_result_new (G_OBJECT (stream),
+						 callback, user_data,
+						 soup_client_input_stream_close_async);
+	cad->priority = priority;
+	cad->cancellable = cancellable ? g_object_ref (cancellable) : NULL;
 
-	if (close_async_ready (cistream->priv->msg, task) == G_SOURCE_CONTINUE) {
-		source = soup_message_io_get_source (cistream->priv->msg,
-						     cancellable, NULL, NULL);
-
-		g_task_attach_source (task, source, (GSourceFunc) close_async_ready);
-		g_source_unref (source);
-	}
+	source = soup_message_io_get_source (cad->cistream->priv->msg,
+					     cancellable,
+					     close_async_ready, cad);
+	g_source_set_priority (source, priority);
+	g_source_attach (source, g_main_context_get_thread_default ());
+	g_source_unref (source);
 }
 
 static gboolean
@@ -205,7 +240,12 @@ soup_client_input_stream_close_finish (GInputStream  *stream,
 				       GAsyncResult  *result,
 				       GError       **error)
 {
-	return g_task_propagate_boolean (G_TASK (result), error);
+	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (result);
+
+	if (g_simple_async_result_propagate_error (simple, error))
+		return FALSE;
+	else
+		return g_simple_async_result_get_op_res_gboolean (simple);
 }
 
 static void
@@ -214,22 +254,23 @@ soup_client_input_stream_class_init (SoupClientInputStreamClass *stream_class)
 	GObjectClass *object_class = G_OBJECT_CLASS (stream_class);
 	GInputStreamClass *input_stream_class = G_INPUT_STREAM_CLASS (stream_class);
 
-	object_class->finalize = soup_client_input_stream_finalize;
-	object_class->set_property = soup_client_input_stream_set_property;
-	object_class->get_property = soup_client_input_stream_get_property;
+	g_type_class_add_private (stream_class, sizeof (SoupClientInputStreamPrivate));
+
+	object_class->set_property = set_property;
+	object_class->get_property = get_property;
 
 	input_stream_class->read_fn = soup_client_input_stream_read_fn;
 	input_stream_class->close_fn = soup_client_input_stream_close_fn;
 	input_stream_class->close_async = soup_client_input_stream_close_async;
 	input_stream_class->close_finish = soup_client_input_stream_close_finish;
 
-	signals[SIGNAL_EOF] =
+	signals[EOF] =
 		g_signal_new ("eof",
 			      G_OBJECT_CLASS_TYPE (object_class),
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      NULL,
+			      _soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
 
 	g_object_class_install_property (
@@ -238,8 +279,7 @@ soup_client_input_stream_class_init (SoupClientInputStreamClass *stream_class)
 				     "Message",
 				     "Message",
 				     SOUP_TYPE_MESSAGE,
-				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
-				     G_PARAM_STATIC_STRINGS));
+				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 }
 
 static void
