@@ -149,8 +149,12 @@ soup_message_io_stop (SoupMessage *msg)
 		io->unpause_source = NULL;
 	}
 
-	if (io->read_state < SOUP_MESSAGE_IO_STATE_FINISHING)
-		g_io_stream_close (io->iostream, NULL, NULL);
+	if (io->read_state < SOUP_MESSAGE_IO_STATE_FINISHING) {
+		if (io->item && io->item->conn)
+			soup_connection_disconnect (io->item->conn);
+		else
+			g_io_stream_close (io->iostream, NULL, NULL);
+	}
 }
 
 void
@@ -686,20 +690,32 @@ io_read (SoupMessage *msg, GCancellable *cancellable, GError **error)
 typedef struct {
 	GSource source;
 	SoupMessage *msg;
+	gboolean paused;
 } SoupMessageSource;
 
+static gboolean
+message_source_check (GSource *source)
+{
+	SoupMessageSource *message_source = (SoupMessageSource *)source;
+
+	if (message_source->paused) {
+		SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (message_source->msg);
+		SoupMessageIOData *io = priv->io_data;
+
+		if (!io || io->paused)
+			return FALSE;
+		else
+			return TRUE;
+	} else
+		return FALSE;
+}
+
 static gboolean
 message_source_prepare (GSource *source,
 			gint    *timeout)
 {
 	*timeout = -1;
-	return FALSE;
-}
-
-static gboolean
-message_source_check (GSource *source)
-{
-	return FALSE;
+	return message_source_check (source);
 }
 
 static gboolean
@@ -707,52 +723,51 @@ message_source_dispatch (GSource     *source,
 			 GSourceFunc  callback,
 			 gpointer     user_data)
 {
-  SoupMessageSourceFunc func = (SoupMessageSourceFunc)callback;
-  SoupMessageSource *message_source = (SoupMessageSource *)source;
+	SoupMessageSourceFunc func = (SoupMessageSourceFunc)callback;
+	SoupMessageSource *message_source = (SoupMessageSource *)source;
 
-  return (*func) (message_source->msg, user_data);
+	return (*func) (message_source->msg, user_data);
 }
 
 static void
 message_source_finalize (GSource *source)
 {
-  SoupMessageSource *message_source = (SoupMessageSource *)source;
+	SoupMessageSource *message_source = (SoupMessageSource *)source;
 
-  g_object_unref (message_source->msg);
+	g_object_unref (message_source->msg);
 }
 
 static gboolean
 message_source_closure_callback (SoupMessage *msg,
 				 gpointer     data)
 {
-  GClosure *closure = data;
+	GClosure *closure = data;
+	GValue param = G_VALUE_INIT;
+	GValue result_value = G_VALUE_INIT;
+	gboolean result;
 
-  GValue param = G_VALUE_INIT;
-  GValue result_value = G_VALUE_INIT;
-  gboolean result;
+	g_value_init (&result_value, G_TYPE_BOOLEAN);
 
-  g_value_init (&result_value, G_TYPE_BOOLEAN);
+	g_value_init (&param, SOUP_TYPE_MESSAGE);
+	g_value_set_object (&param, msg);
 
-  g_value_init (&param, SOUP_TYPE_MESSAGE);
-  g_value_set_object (&param, msg);
+	g_closure_invoke (closure, &result_value, 1, &param, NULL);
 
-  g_closure_invoke (closure, &result_value, 1, &param, NULL);
+	result = g_value_get_boolean (&result_value);
+	g_value_unset (&result_value);
+	g_value_unset (&param);
 
-  result = g_value_get_boolean (&result_value);
-  g_value_unset (&result_value);
-  g_value_unset (&param);
-
-  return result;
+	return result;
 }
 
 static GSourceFuncs message_source_funcs =
 {
-  message_source_prepare,
-  message_source_check,
-  message_source_dispatch,
-  message_source_finalize,
-  (GSourceFunc)message_source_closure_callback,
-  (GSourceDummyMarshal)g_cclosure_marshal_generic,
+	message_source_prepare,
+	message_source_check,
+	message_source_dispatch,
+	message_source_finalize,
+	(GSourceFunc)message_source_closure_callback,
+	(GSourceDummyMarshal)g_cclosure_marshal_generic,
 };
 
 GSource *
@@ -764,7 +779,11 @@ soup_message_io_get_source (SoupMessage *msg, GCancellable *cancellable,
 	GSource *base_source, *source;
 	SoupMessageSource *message_source;
 
-	if (io && SOUP_MESSAGE_IO_STATE_POLLABLE (io->read_state)) {
+	if (!io) {
+		base_source = g_timeout_source_new (0);
+	} else if (io->paused) {
+		base_source = NULL;
+	} else if (SOUP_MESSAGE_IO_STATE_POLLABLE (io->read_state)) {
 		GPollableInputStream *istream;
 
 		if (io->body_istream)
@@ -772,7 +791,7 @@ soup_message_io_get_source (SoupMessage *msg, GCancellable *cancellable,
 		else
 			istream = G_POLLABLE_INPUT_STREAM (io->istream);
 		base_source = g_pollable_input_stream_create_source (istream, cancellable);
-	} else if (io && SOUP_MESSAGE_IO_STATE_POLLABLE (io->write_state)) {
+	} else if (SOUP_MESSAGE_IO_STATE_POLLABLE (io->write_state)) {
 		GPollableOutputStream *ostream;
 
 		if (io->body_ostream)
@@ -783,15 +802,18 @@ soup_message_io_get_source (SoupMessage *msg, GCancellable *cancellable,
 	} else
 		base_source = g_timeout_source_new (0);
 
-	g_source_set_dummy_callback (base_source);
 	source = g_source_new (&message_source_funcs,
 			       sizeof (SoupMessageSource));
 	g_source_set_name (source, "SoupMessageSource");
 	message_source = (SoupMessageSource *)source;
 	message_source->msg = g_object_ref (msg);
+	message_source->paused = io && io->paused;
 
-	g_source_add_child_source (source, base_source);
-	g_source_unref (base_source);
+	if (base_source) {
+		g_source_set_dummy_callback (base_source);
+		g_source_add_child_source (source, base_source);
+		g_source_unref (base_source);
+	}
 	g_source_set_callback (source, (GSourceFunc) callback, user_data, NULL);
 	return source;
 }
@@ -804,6 +826,7 @@ io_run_until (SoupMessage *msg,
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupMessageIOData *io = priv->io_data;
 	gboolean progress = TRUE, done;
+	GError *my_error = NULL;
 
 	if (g_cancellable_set_error_if_cancelled (cancellable, error))
 		return FALSE;
@@ -820,17 +843,39 @@ io_run_until (SoupMessage *msg,
 	       (io->read_state < read_state || io->write_state < write_state)) {
 
 		if (SOUP_MESSAGE_IO_STATE_ACTIVE (io->read_state))
-			progress = io_read (msg, cancellable, error);
+			progress = io_read (msg, cancellable, &my_error);
 		else if (SOUP_MESSAGE_IO_STATE_ACTIVE (io->write_state))
-			progress = io_write (msg, cancellable, error);
+			progress = io_write (msg, cancellable, &my_error);
 		else
 			progress = FALSE;
 	}
 
-	done = (priv->io_data == io &&
-		io->read_state >= read_state &&
+	if (my_error) {
+		g_propagate_error (error, my_error);
+		g_object_unref (msg);
+		return FALSE;
+	} else if (g_cancellable_set_error_if_cancelled (cancellable, error)) {
+		g_object_unref (msg);
+		return FALSE;
+	} else if (priv->io_data != io) {
+		g_set_error_literal (error, G_IO_ERROR,
+				     G_IO_ERROR_CANCELLED,
+				     _("Operation was cancelled"));
+		g_object_unref (msg);
+		return FALSE;
+	}
+
+	done = (io->read_state >= read_state &&
 		io->write_state >= write_state);
 
+	if (io->paused && !done) {
+		g_set_error_literal (error, G_IO_ERROR,
+				     G_IO_ERROR_WOULD_BLOCK,
+				     _("Operation would block"));
+		g_object_unref (msg);
+		return FALSE;
+	}
+
 	g_object_unref (msg);
 	return done;
 }
@@ -841,6 +886,7 @@ io_run (SoupMessage *msg, gpointer user_data)
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupMessageIOData *io = priv->io_data;
 	GError *error = NULL;
+	GCancellable *cancellable;
 
 	if (io->io_source) {
 		g_source_destroy (io->io_source);
@@ -849,21 +895,24 @@ io_run (SoupMessage *msg, gpointer user_data)
 	}
 
 	g_object_ref (msg);
+	cancellable = io->cancellable ? g_object_ref (io->cancellable) : NULL;
 
 	if (io_run_until (msg,
 			  SOUP_MESSAGE_IO_STATE_DONE,
 			  SOUP_MESSAGE_IO_STATE_DONE,
-			  io->cancellable, &error)) {
+			  cancellable, &error)) {
 		soup_message_io_finished (msg);
 	} else if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
 		g_clear_error (&error);
 		io->io_source = soup_message_io_get_source (msg, NULL, io_run, msg);
 		g_source_attach (io->io_source, io->async_context);
-	} else if (error) {
+	} else if (error && priv->io_data == io) {
 		io_error (msg, error);
 	}
 
 	g_object_unref (msg);
+	g_clear_object (&cancellable);
+
 	return FALSE;
 }
 
@@ -1042,6 +1091,9 @@ soup_message_io_pause (SoupMessage *msg)
 
 	g_return_if_fail (io != NULL);
 
+	if (io->item && io->item->new_api)
+		g_return_if_fail (io->read_state < SOUP_MESSAGE_IO_STATE_BODY);
+
 	if (io->io_source) {
 		g_source_destroy (io->io_source);
 		g_source_unref (io->io_source);
@@ -1081,6 +1133,12 @@ soup_message_io_unpause (SoupMessage *msg)
 
 	g_return_if_fail (io != NULL);
 
+	if (io->item && io->item->new_api) {
+		g_return_if_fail (io->read_state < SOUP_MESSAGE_IO_STATE_BODY);
+		io->paused = FALSE;
+		return;
+	}
+
 	if (!io->blocking) {
 		if (!io->unpause_source) {
 			io->unpause_source = soup_add_completion (
