@@ -162,7 +162,7 @@ conditional_get_ready_cb (SoupSession *session, SoupMessage *msg, gpointer user_
 			/* FIXME: this is wrong; the cache won't have
 			 * the sniffed type.
 			 */
-			sadata->http->priv->content_type = g_strdup (soup_message_headers_get_content_type (sadata->original->response_headers, NULL));
+			sadata->http->priv->content_type = g_strdup (soup_message_headers_get_content_type (msg->response_headers, NULL));
 
 			g_simple_async_result_complete (sadata->simple);
 
@@ -172,10 +172,12 @@ conditional_get_ready_cb (SoupSession *session, SoupMessage *msg, gpointer user_
 		}
 	}
 
-	/* The resource was modified, or else it mysteriously disappeared
-	 * from our cache. Either way we need to reload it now.
+	/* The resource was modified or the server returned a 200
+	 * OK. Either way we reload it. This is far from optimal as
+	 * we're donwloading the resource twice, but we will change it
+	 * once the cache is integrated in the streams stack.
 	 */
-	soup_session_send_request_async (session, msg, sadata->cancellable,
+	soup_session_send_request_async (session, sadata->original, sadata->cancellable,
 					 http_input_stream_ready_cb, sadata);
 }
 
@@ -344,7 +346,7 @@ soup_request_http_class_init (SoupRequestHTTPClass *request_http_class)
  *
  * Gets a new reference to the #SoupMessage associated to this SoupRequest
  *
- * Returns: a new reference to the #SoupMessage
+ * Returns: (transfer full): a new reference to the #SoupMessage
  *
  * Since: 2.34
  */
